def run(action, music=None, fade_out=0, wait=False,
        set_wait_id=None, **kwargs):
    previous = None
    for music in action.music_list(music):
        if music.is_loaded_paused() or music.is_loaded_playing():
            if previous is not None:
                previous.stop(fade_out=fade_out)
            previous = music
        else:
            music.stop(fade_out=fade_out)

    if previous is not None:
        action.waiting_music = previous
        previous.stop(
                fade_out=fade_out,
                wait=wait,
                set_wait_id=set_wait_id)

def description(action, music=None, fade_out=0, wait=False,
        set_wait_id=None, **kwargs):

    formats = []
    message = "stopping "
    if music is not None:
        message += "music « {} »"
        formats.append(music.name)
    else:
        message += "all musics"

    if fade_out > 0:
        message += " with {}s fadeout"
        formats.append(fade_out)
        if wait:
            if set_wait_id is not None:
                message += " (waiting the end of fadeout, with id {})"
                formats.append(set_wait_id)
            else:
                message += " (waiting the end of fadeout)"

    return _(message).format(*formats)

def interrupt(action, music=None, fade_out=0, wait=False,
        set_wait_id=None, **kwargs):
    if action.waiting_music is not None:
        action.waiting_music.wait_event.set()



import markdown
from markdown.extensions.toc import slugify

def lang_slugify(lang):
    def l_slugify(value, separator):
        return slugify(lang + "__" + value, separator)
    return l_slugify

def get_markdown(md_file, lang, table_of_content):
    with open(md_file, "r") as f:
        text = f.read()

    return markdown.markdown(
        text=text,
        extensions=[
            'markdown.extensions.codehilite',
            'markdown.extensions.toc',
            'markdown.extensions.tables'
        ],
        extension_configs={
            'markdown.extensions.codehilite': {
                'noclasses': True,
            },
            'markdown.extensions.toc': {
                'title': table_of_content,
                'slugify': lang_slugify(lang)
            }
        })


print("<p><a href='#lang-en'>Documenation in english</a></p>")
print("<p><a href='#lang-fr'>Documenation en Français</a></p>")
print("<p><a href='#lang-bottom'>Bottom of documentation</a></p>")
print("<hr id='lang-en'/>")
print(get_markdown("documentation_en.md", "en", "Table of contents"))
print("<hr id='lang-fr'/>")
print(get_markdown("documentation_fr.md", "fr", "Sommaire"))
print("<hr id='lang-bottom' style='visibility: hidden' />")



def run(action, music=None, **kwargs):
    for music in action.music_list(music):
        if music.is_loaded_playing():
            music.pause()

def description(action, music=None, **kwargs):
    if music is not None:
        return _("pausing « {} »").format(music.name)
    else:
        return _("pausing all musics")



from . import interrupt_wait
from . import pause
from . import pause_wait
from . import play
from . import reset_wait
from . import run_command
from . import seek
from . import stop
from . import stop_all_actions
from . import unpause
from . import unpause_wait
from . import volume
from . import wait



def run(action, music=None, fade_in=0, start_at=0,
        restart_if_running=False, volume=100,
        loop=0, **kwargs):
    for music in action.music_list(music):
        if restart_if_running:
            if music.is_in_use():
                music.stop()
            music.play(
                    volume=volume,
                    fade_in=fade_in,
                    start_at=start_at,
                    loop=loop)
        elif not music.is_in_use():
            music.play(
                    volume=volume,
                    fade_in=fade_in,
                    start_at=start_at,
                    loop=loop)

def description(action, music=None, fade_in=0, start_at=0,
        restart_if_running=False, volume=100, loop=0, **kwargs):
    formats = []
    message = "starting "
    if music is not None:
        message += "« {} »"
        formats.append(music.name)
    else:
        message += "all musics"

    if start_at != 0:
        message += " at {}s"
        formats.append(start_at)

    if fade_in != 0:
        message += " with {}s fade_in"
        formats.append(fade_in)

    message += " at volume {}%"
    formats.append(volume)

    if loop > 0:
        message += " {} times"
        formats.append(loop + 1)
    elif loop < 0:
        message += " in loop"

    if restart_if_running:
        message += " (restarting if already running)"

    return _(message).format(*formats)



def run(action, wait_id=None, **kwargs):
    action.mapping.reset_wait(wait_id)

def description(action, wait_id=None, **kwargs):
    if wait_id is None:
        return _("reset all waits")
    else:
        return _("reset wait with id {}").format(wait_id)



from kivy.uix.widget import Widget
from kivy.properties import AliasProperty, BooleanProperty, \
                            ListProperty, StringProperty
from kivy.uix.behaviors import ButtonBehavior

from .action import Action
from .helpers import debug_print
import time
import threading
from transitions.extensions import HierarchicalMachine as Machine

class KeyMachine(Widget):
    STATES = [
        'initial',
        'configuring',
        'configured',
        'loading',
        'failed',
        {
            'name': 'loaded',
            'children': [
                'no_config',
                'no_actions',
                'running',
                'protecting_repeat'
            ]
        }
    ]

    TRANSITIONS = [
        {
            'trigger': 'configure',
            'source': 'initial',
            'dest': 'configuring'
        },
        {
            'trigger': 'fail',
            'source': 'configuring',
            'dest': 'failed',
            'after': 'key_loaded_callback'
        },
        {
            'trigger': 'success',
            'source': 'configuring',
            'dest': 'configured',
            'after': 'load'
        },
        {
            'trigger': 'no_config',
            'source': 'configuring',
            'dest': 'loaded_no_config',
            'after': 'key_loaded_callback'
        },
        {
            'trigger': 'load',
            'source': 'configured',
            'dest': 'loading'
        },
        {
            'trigger': 'fail',
            'source': 'loading',
            'dest': 'failed',
            'after': 'key_loaded_callback'
        },
        {
            'trigger': 'success',
            'source': 'loading',
            'dest': 'loaded',
            'after': 'key_loaded_callback'
        },
        {
            'trigger': 'no_actions',
            'source': 'loading',
            'dest': 'loaded_no_actions',
            'after': 'key_loaded_callback'
        },
        {
            'trigger': 'reload',
            'source': ['loaded','failed'],
            'dest': 'configuring',
            'after': 'key_loaded_callback'
        },
        {
            'trigger': 'run',
            'source': 'loaded',
            'dest': 'loaded_running',
            'after': ['run_actions', 'finish'],
            # if a child, like loaded_no_actions, has no transitions, then it
            # is bubbled to the parent, and we don't want that.
            'conditions': ['is_loaded']
        },
        {
            'trigger': 'finish',
            'source': 'loaded_running',
            'dest': 'loaded_protecting_repeat'
        },
        {
            'trigger': 'repeat_protection_finished',
            'source': 'loaded_protecting_repeat',
            'dest': 'loaded'
        },
    ]

    state = StringProperty("")

    def __init__(self, key, **kwargs):
        self.key = key

        Machine(model=self, states=self.STATES,
                transitions=self.TRANSITIONS, initial='initial',
                ignore_invalid_triggers=True, queued=True)
        super(KeyMachine, self).__init__(**kwargs)

    # Machine states / events
    def is_loaded_or_failed(self):
        return self.is_loaded(allow_substates=True) or self.is_failed()

    def is_loaded_inactive(self):
        return self.is_loaded_no_config() or self.is_loaded_no_actions()

    def on_enter_configuring(self):
        if self.key.key_sym in self.key.parent.key_config:
            self.key.config = self.key.parent.key_config[self.key.key_sym]

            self.key.actions = []
            for key_action in self.key.config['actions']:
                self.key.add_action(key_action[0], **key_action[1])

            if 'description' in self.key.config['properties']:
                self.key.set_description(self.key.config['properties']['description'])
            else:
                self.key.unset_description()
            if 'color' in self.key.config['properties']:
                self.key.set_color(self.key.config['properties']['color'])
            else:
                self.key.unset_color()
            self.success()
        else:
            self.no_config()

    def on_enter_loading(self):
        if len(self.key.actions) > 0:
            for action in self.key.actions:
                action.load()
        else:
            self.no_actions()

    def run_actions(self, modifiers):
        self.key.parent.parent.ids['KeyList'].append(self.key.key_sym)
        debug_print("running actions for {}".format(self.key.key_sym))
        start_time = time.time()
        self.key.parent.start_running(self.key, start_time)
        for self.key.current_action in self.key.actions:
            if self.key.parent.keep_running(self.key, start_time):
                self.key.list_actions()
                self.key.current_action.run(start_time)
        self.key.list_actions(last_action_finished=True)

        self.key.parent.finished_running(self.key, start_time)

    def on_enter_loaded_protecting_repeat(self, modifiers):
        if self.key.repeat_delay > 0:
            self.key.protecting_repeat_timer = threading.Timer(
                    self.key.repeat_delay,
                    self.key.repeat_protection_finished)
            self.key.protecting_repeat_timer.start()
        else:
            self.key.repeat_protection_finished()

    # Callbacks
    def key_loaded_callback(self):
        self.key.parent.key_loaded_callback()


class Key(ButtonBehavior, Widget):

    key_sym = StringProperty(None)
    custom_color = ListProperty([0, 1, 0])
    description_title = StringProperty("")
    description = ListProperty([])
    machine_state = StringProperty("")

    def get_alias_line_cross_color(self):
        if not self.is_failed() and (
                not self.is_loaded(allow_substates=True)\
                or self.is_loaded_running()\
                or self.is_loaded_protecting_repeat()):
            return [120/255, 120/255, 120/255, 1]
        else:
            return [0, 0, 0, 0]

    def set_alias_line_cross_color(self):
        pass

    line_cross_color = AliasProperty(
            get_alias_line_cross_color,
            set_alias_line_cross_color,
            bind=['machine_state'])

    def get_alias_line_color(self):
        if self.is_loaded_running():
            return [0, 0, 0, 1]
        else:
            return [120/255, 120/255, 120/255, 1]

    def set_alias_line_color(self):
        pass

    line_color = AliasProperty(get_alias_line_color, set_alias_line_color,
            bind=['machine_state'])

    def get_alias_color(self):
        if self.is_loaded_inactive():
            return [1, 1, 1, 1]
        elif self.is_loaded_protecting_repeat():
            return [*self.custom_color, 100/255]
        elif self.is_loaded_running():
            return [*self.custom_color, 100/255]
        elif self.is_loaded(allow_substates=True):
            return [*self.custom_color, 1]
        elif self.is_failed():
            return [0, 0, 0, 1]
        else:
            return [*self.custom_color, 100/255]
    def set_alias_color(self):
        pass

    color = AliasProperty(get_alias_color, set_alias_color,
            bind=['machine_state', 'custom_color'])

    def __getattr__(self, name):
        if hasattr(self.machine, name):
            return getattr(self.machine, name)
        else:
            raise AttributeError

    def machine_state_changed(self, instance, machine_state):
        self.machine_state = self.machine.state

    def __init__(self, **kwargs):
        self.actions = []
        self.current_action = None
        self.machine = KeyMachine(self)
        self.machine.bind(state=self.machine_state_changed)

        super(Key, self).__init__(**kwargs)

    # Kivy events
    def on_key_sym(self, key, key_sym):
        if key_sym != "":
            self.configure()

    def on_press(self):
        self.list_actions()

    # This one cannot be in the Machine state since it would be queued to run
    # *after* the loop is ended...
    def interrupt(self):
        self.current_action.interrupt()

    # Callbacks
    def callback_action_ready(self, action, success):
        if not success:
            self.fail()
        elif all(action.is_loaded_or_failed() for action in self.actions):
            self.success()

    # Setters
    def set_description(self, description):
        if description[0] is not None:
            self.description_title = str(description[0])
        self.description = []
        for desc in description[1 :]:
            if desc is None:
                self.description.append("")
            else:
                self.description.append(str(desc).replace(" ", " "))

    def unset_description(self):
        self.description_title = ""
        self.description = []

    def set_color(self, color):
        color = [x / 255 for x in color]
        self.custom_color = color

    def unset_color(self):
        self.custom_color = [0, 1, 0]

    # Helpers
    @property
    def repeat_delay(self):
         if hasattr(self, 'config') and\
                 'repeat_delay' in self.config['properties']:
             return self.config['properties']['repeat_delay']
         else:
             return 0

    # Actions handling
    def add_action(self, action_name, **arguments):
        self.actions.append(Action(action_name, self, **arguments))

    def list_actions(self, last_action_finished=False):
        not_running = (not self.is_loaded_running())
        current_action_seen = False
        action_descriptions = []
        for action in self.actions:
            if not_running:
                state = "inactive"
            elif last_action_finished:
                state = "done"
            elif current_action_seen:
                state = "pending"
            elif action == self.current_action:
                current_action_seen = True
                state = "current"
            else:
                state = "done"
            action_descriptions.append([action.description(), state])
        self.parent.parent.ids['ActionList'].update_list(
                self,
                action_descriptions)



def run(action, music=None, value=100, fade=0, delta=False, **kwargs):
    if music is not None:
        music.set_volume(value, delta=delta, fade=fade)
    else:
        action.mapping.set_master_volume(value, delta=delta, fade=fade)

def description(action, music=None,
        value=100, delta=False, fade=0, **kwargs):
    formats = []
    message = ""
    if delta:
        if music is not None:
            message += "{:+d}% to volume of « {} »"
            formats.append(value)
            formats.append(music.name)
        else:
            message += "{:+d}% to volume"
            formats.append(value)
    else:
        if music is not None:
            message += "setting volume of « {} » to {}%"
            formats.append(music.name)
            formats.append(value)
        else:
            message += "setting volume to {}%"
            formats.append(value)

    if fade > 0:
        message += " with {}s fade"
        formats.append(fade)

    return _(message).format(*formats)



import threading
import time

def run(action, duration=0, music=None, set_wait_id=None, **kwargs):
    action.mapping.add_wait(action, wait_id=set_wait_id)

    action.sleep_event = threading.Event()
    action.sleep_event_timer = threading.Timer(
            duration,
            action.sleep_event.set)

    action.sleep_event_initial_duration = duration
    action.sleep_event_paused = False
    action.sleep_event_left_time = duration

    if music is not None:
        music.wait_end()

    if duration <= 0 or not action.sleep_event_paused:
        action.sleep_event_timer.start()
        action.sleep_event_started_time = time.time()

    action.sleep_event.wait()

def description(action, duration=0, music=None, set_wait_id=None, **kwargs):
    formats = []
    message = ""
    if music is None:
        message += "waiting {}s"
        formats.append(duration)
    elif duration == 0:
        message += "waiting the end of « {} »"
        formats.append(music.name)
    else:
        message += "waiting the end of « {} » + {}s"
        formats.append(music.name)
        formats.append(duration)

    if set_wait_id is not None:
        message += " (setting id = {})"
        formats.append(set_wait_id)

    return _(message).format(*formats)

def pause(action, **kwargs):
    if action.sleep_event_paused:
        return

    action.sleep_event_paused = True

    if not action.sleep_event_timer.is_alive():
        return

    action.sleep_event_timer.cancel()

    action.sleep_event_left_time = action.sleep_event_left_time\
            - (time.time() - action.sleep_event_started_time)
    if action.sleep_event_left_time < 0:
        action.sleep_event.set()

def unpause(action, **kwargs):
    if not action.sleep_event_paused:
        return

    action.sleep_event_paused = False

    action.sleep_event_timer = threading.Timer(
            action.sleep_event_left_time,
            action.sleep_event.set)

    action.sleep_event_timer.start()
    action.sleep_event_started_time = time.time()

def reset(action, **kwargs):
    action.sleep_event_timer.cancel()

    action.sleep_event_left_time = action.sleep_event_initial_duration

    if action.sleep_event_paused:
        return

    action.sleep_event_timer = threading.Timer(
            action.sleep_event_left_time,
            action.sleep_event.set)

    action.sleep_event_timer.start()
    action.sleep_event_started_time = time.time()

def interrupt(action, duration=0, music=None, **kwargs):
    if action.sleep_event is not None:
        action.sleep_event.set()
        action.sleep_event_timer.cancel()
    if music is not None:
        music.wait_event.set()



import threading

from .helpers import debug_print

class Lock:
    def __init__(self, lock_type):
        self.type = lock_type
        self.lock = threading.RLock()

    def __enter__(self, *args, **kwargs):
        self.acquire(*args, **kwargs)

    def __exit__(self, type, value, traceback, *args, **kwargs):
        self.release(*args, **kwargs)

    def acquire(self, *args, **kwargs):
        #debug_print("acquiring lock for {}".format(self.type))
        self.lock.acquire(*args, **kwargs)

    def release(self, *args, **kwargs):
        #debug_print("releasing lock for {}".format(self.type))
        self.lock.release(*args, **kwargs)




# -*- coding: utf-8 -*-
import argparse
import sys
import os
import math
import sounddevice as sd
import logging
import gettext
import yaml
gettext.install('music_sampler')
Logger = logging.getLogger("kivy")

from . import sysfont

class Config:
    pass

def find_font(name, style=sysfont.STYLE_NONE):
    if getattr(sys, 'frozen', False):
        font = sys._MEIPASS + "/fonts/{}_{}.ttf".format(name, style)
    else:
        font = sysfont.get_font(name, style=style)
        if font is not None:
            font = font[4]
    return font

def register_fonts():
    from kivy.core.text import LabelBase

    ubuntu_regular = find_font("Ubuntu", style=sysfont.STYLE_NORMAL)
    ubuntu_bold = find_font("Ubuntu", style=sysfont.STYLE_BOLD)
    symbola = find_font("Symbola")

    if ubuntu_regular is None:
        error_print("Font Ubuntu regular could not be found, "
                "please install it.", exit=True)
    if symbola is None:
        error_print("Font Symbola could not be found, please install it.",
                exit=True)
    if ubuntu_bold is None:
        warn_print("Font Ubuntu Bold could not be found.")

    LabelBase.register(name="Ubuntu",
            fn_regular=ubuntu_regular,
            fn_bold=ubuntu_bold)
    LabelBase.register(name="Symbola",
            fn_regular=symbola)


def path():
    if getattr(sys, 'frozen', False):
        return sys._MEIPASS + "/"
    else:
        return os.path.dirname(os.path.realpath(__file__))


Configs = {
    'music_path': {
        'abbr': '-p',
        'default': '.',
        'help': _("Folder in which to find the music files"),
        'type': None
    },
    'latency': {
        'abbr': '-l',
        'default': 'high',
        'help': _("Latency: low, high or number of seconds"),
        'type': None
    },
    'language': {
        'abbr': '-L',
        'default': "fr",
        'help': _("Select another language"),
        'type': None
    },
    'device': {
        'abbr': '-d',
        'default': None,
        'help': _("Select this sound device"),
        'type': None
    },
    'blocksize': {
        'abbr': '-b',
        'default': 0,
        'help': _("Blocksize: If not 0, the number of frames to take\
                    at each step for the mixer"),
        'type': int
    },
    'frame_rate': {
        'abbr': '-f',
        'default': 44100,
        'help': _("Frame rate to play the musics"),
        'type': int
    },
    'channels': {
        'abbr': '-x',
        'default': 2,
        'help': _("Number of channels to use"),
        'type': int
    },
    'sample_width': {
        'abbr': '-s',
        'default': 2,
        'help': _("Sample width (number of bytes for each frame)"),
        'type': int
    },
    'builtin_mixing': {
        'default': False,
        'help_yes': _("Make the mixing of sounds manually\
                    (do it if the system cannot handle it correctly)"),
        'help_no': _("Don't make the mixing of sounds manually (default)"),
        'type': 'boolean'
    },
    'debug': {
        'abbr': '-d',
        'default': False,
        'help_yes': _("Print messages in console"),
        'help_no': _("Don't print messages in console (default)"),
        'type': 'boolean'
    },
    'focus_warning': {
        'default': True,
        'help_yes': _("Show a warning when focus is lost (default)"),
        'help_no': _("Don't show warning when focus is lost"),
        'type': 'boolean'
    },
    'list_devices': {
        'help': _("List available sound devices"),
        'type': 'action'
    },
}
Configs_order = [
    'debug',
    'music_path',
    'builtin_mixing',
    'latency',
    'blocksize',
    'frame_rate',
    'channels',
    'sample_width',
    'focus_warning',
    'language',
    'list_devices',
    'device',
]
def parse_args():
    argv = sys.argv[1 :]
    sys.argv = sys.argv[: 1]
    if "--" in argv:
        index = argv.index("--")
        kivy_args = argv[index+1 :]
        argv = argv[: index]

        sys.argv.extend(kivy_args)

    os.environ["KIVY_NO_CONFIG"] = 'true'
    sys.argv.extend(["-c", "kivy:log_level:warning"])
    sys.argv.extend(["-c", "kivy:log_dir:/tmp"])
    sys.argv.extend(["-c", "kivy:log_name:/tmp/music_sampler_%_.txt"])

    parser = argparse.ArgumentParser(
            argument_default=argparse.SUPPRESS,
            description=_("A Music Sampler application."))
    parser.add_argument("-V", "--version",
            action="version",
            help=_("Displays the current version and exits. Only use\
                    in bundled package"),
            version=show_version())
    parser.add_argument("-c", "--config",
            default="config.yml",
            required=False,
            help=_("Config file to load (default: config.yml)"))
    for argument in Configs_order:
        arg = Configs[argument]
        if arg['type'] != 'boolean' and arg['type'] != 'action':
            parser.add_argument(arg['abbr'], '--' + argument.replace('_', '-'),
                    type=arg['type'],
                    help=arg['help']+_(" (default: {})").format(arg['default']))
        elif arg['type'] == 'boolean':
            parser.add_argument('--' + argument.replace('_', '-'),
                    action='store_const', const=True,
                    help=arg['help_yes'])
            parser.add_argument('--no-' + argument.replace('_', '-'),
                    action='store_const', const=True,
                    help=arg['help_no'])
        else:
            parser.add_argument('--' + argument.replace('_', '-'),
                    action='store_const', const=True,
                    help=arg['help'])
    parser.add_argument('--',
            dest="args",
            help=_("Kivy arguments. All arguments after this are interpreted\
                    by Kivy. Pass \"-- --help\" to get Kivy's usage."))

    args = parser.parse_args(argv)

    Config.yml_file = args.config
    build_config(args)

    if Config.device is not None:
        sd.default.device = Config.device

    if Config.list_devices:
        print(sd.query_devices())
        sys.exit()

    if Config.debug:
        sys.argv.extend(["-c", "kivy:log_level:debug"])

    if Config.language != 'en':
        gettext.translation("music_sampler",
                localedir=path() + '/locales',
                languages=[Config.language]).install()
    if not Config.music_path.endswith("/"):
        Config.music_path = Config.music_path + "/"

def dump_config():
    max_size = max(max(map(len, Configs_order)), len('config'))
    info_print("{:<{}} : {}".format(
        "config", max_size, Config.yml_file))
    for item in Config.__dict__:
        if item in Configs_order:
            info_print("{:<{}} : {}".format(
                item, max_size, getattr(Config, item)))

def build_config(args):
    stream = open(Config.yml_file, "r")
    try:
        config = yaml.safe_load(stream)
    except Exception as e:
        error_print("Error while loading config file: {}".format(e))
        config = {}
    stream.close()
    if 'config' in config:
        config = config['config']
    else:
        config = {}

    for config_item in Configs_order:
        if Configs[config_item]['type'] != 'boolean' and \
                Configs[config_item]['type'] != 'action':
            t = Configs[config_item]['type'] or str
            if hasattr(args, config_item):
                setattr(Config, config_item, getattr(args, config_item))
            elif config_item in config:
                setattr(Config, config_item, t(config[config_item]))
            else:
                setattr(Config, config_item, Configs[config_item]['default'])
        elif Configs[config_item]['type'] == 'boolean':
            if hasattr(args, 'no_' + config_item) or hasattr(args, config_item):
                setattr(Config, config_item, hasattr(args, config_item))
            elif config_item in config:
                setattr(Config, config_item, config[config_item])
            else:
                setattr(Config, config_item, Configs[config_item]['default'])
        else:
            setattr(Config, config_item, hasattr(args, config_item))


def show_version():
    if getattr(sys, 'frozen', False):
        with open(path() + ".pyinstaller_commit", "r") as f:
            return f.read()
    else:
        return _("option '-V' can only be used in bundled package")

def duration_to_min_sec(duration):
    minutes = int(duration / 60)
    seconds = int(duration) % 60
    if minutes < 100:
        return "{:2}:{:0>2}".format(minutes, seconds)
    else:
        return "{}:{:0>2}".format(minutes, seconds)

def gain(volume, old_volume=None):
    if old_volume is None:
        return 20 * math.log10(max(volume, 0.1) / 100)
    else:
        return [
                20 * math.log10(max(volume, 0.1) / max(old_volume, 0.1)),
                max(volume, 0)]

def debug_print(message, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    Logger.debug('MusicSampler: ' + message, exc_info=with_trace)

def error_print(message, exit=False, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    # FIXME: handle it correctly when in a thread
    if exit:
        Logger.critical('MusicSampler: ' + message, exc_info=with_trace)
        sys.exit(1)
    else:
        Logger.error('MusicSampler: ' + message, exc_info=with_trace)

def warn_print(message, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    Logger.warn('MusicSampler: ' + message, exc_info=with_trace)

def info_print(message, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    Logger.info('MusicSampler: ' + message, exc_info=with_trace)




from setuptools import setup

def readme():
    with open('README') as f:
        return f.read()

setup(name='music_sampler',
        setup_requires=['setuptools_scm'],
        use_scm_version=True,
        description='A music player which associates each key on the keyboard '
            'to a set of actions to run',
        long_description=readme(),
        classifiers= [
            'Development Status :: 4 - Beta',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3.5',
            'Topic :: Multimedia :: Sound/Audio :: Players'
        ],
        keywords='music sampler keyboard',
        url='https://git.immae.eu/?p=perso/Immae/Projets/Python/MusicSampler.git',
        author='Ismaël Bouya',
        author_email='ismael.bouya@normalesup.org',
        license='MIT',
        packages=['music_sampler', 'music_sampler.actions'],
        install_requires=[
            'Cython>=0.24',
            'Kivy>=1.9.1',
            'pydub>=0.16.4',
            'Pygame>=1.9.2.dev1',
            'sounddevice>=0.3.3',
            'TRANSITIONS>=0.4.1',
            'PyYAML>=3.11'
        ],
        entry_points={
            'console_scripts': ['music_sampler=music_sampler.app:main'],
        },
        include_package_data=True,
        zip_safe=False)




class GainEffect:
    effect_types = [
        'noop',
        'fade'
    ]

    def __init__(self, effect, audio_segment, initial_loop, start, end,
            **kwargs):
        if effect in self.effect_types:
            self.effect = effect
        else:
            self.effect = 'noop'

        self.start = start
        self.end = end
        self.audio_segment = audio_segment
        self.initial_loop = initial_loop
        getattr(self, self.effect + "_init")(**kwargs)

    def get_last_gain(self):
        return getattr(self, self.effect + "_get_last_gain")()

    def get_next_gain(self, current_frame, current_loop, frame_count):
        # This returns two values:
        # - The first one is the gain to apply on that frame
        # - The last one is True or False depending on whether it is the last
        #   call to the function and the last gain should be saved permanently
        return getattr(self, self.effect + "_get_next_gain")(
                current_frame,
                current_loop,
                frame_count)

    # Noop
    def noop_init(self, **kwargs):
        pass

    def noop_get_last_gain(self, **kwargs):
        return 0

    def noop_get_next_gain(self, **kwargs):
        return [0, True]

    # Fading
    def fade_init(self, gain=0, **kwargs):
        self.audio_segment_frame_count = self.audio_segment.frame_count()
        self.first_frame = int(
                self.audio_segment_frame_count * self.initial_loop +\
                self.audio_segment.frame_rate * self.start)
        self.last_frame = int(
                self.audio_segment_frame_count * self.initial_loop +\
                self.audio_segment.frame_rate * self.end)
        self.gain= gain

    def fade_get_last_gain(self):
        return self.gain

    def fade_get_next_gain(self, current_frame, current_loop, frame_count):
        current_frame = current_frame \
                + (current_loop - self.initial_loop) \
                    * self.audio_segment_frame_count

        if current_frame >= self.last_frame:
            return [self.gain, True]
        elif current_frame < self.first_frame:
            return [0, False]
        else:
            return [
                    (current_frame - self.first_frame) / \
                            (self.last_frame - self.first_frame) * self.gain,
                    False
                    ]





import shlex, subprocess

def run(action, command="", wait=False, **kwargs):
    action.process = subprocess.Popen(command, shell=True)
    if wait:
        action.process.wait()

def description(action, command="", wait=False, **kwargs):
    formats = []
    message = "running command {}"
    formats.append(command)
    if wait:
        message += " (waiting for its execution to finish)"

    return _(message).format(*formats)



import threading
import pydub
import time
from transitions.extensions import HierarchicalMachine as Machine

import os.path

import audioop

from .lock import Lock
from .helpers import Config, gain, debug_print, error_print
from .mixer import Mixer
from .music_effect import GainEffect

file_lock = Lock("file")

class MusicFile:
    STATES = [
        'initial',
        'loading',
        'failed',
        {
            'name': 'loaded',
            'children': [
                'playing',
                'paused',
                'stopping'
            ]
        }
    ]
    TRANSITIONS = [
        {
            'trigger': 'load',
            'source': 'initial',
            'dest': 'loading',
            'after': 'poll_loaded'
        },
        {
            'trigger': 'fail',
            'source': 'loading',
            'dest': 'failed'
        },
        {
            'trigger': 'success',
            'source': 'loading',
            'dest': 'loaded'
        },
        {
            'trigger': 'start_playing',
            'source': 'loaded',
            'dest': 'loaded_playing',
            # if a child has no transitions, then it is bubbled to the parent,
            # and we don't want that. Not useful in that machine precisely.
            'conditions': ['is_loaded']
        },
        {
            'trigger': 'pause',
            'source': 'loaded_playing',
            'dest': 'loaded_paused'
        },
        {
            'trigger': 'unpause',
            'source': 'loaded_paused',
            'dest': 'loaded_playing'
        },
        {
            'trigger': 'stop_playing',
            'source': ['loaded_playing','loaded_paused'],
            'dest': 'loaded_stopping'
        },
        {
            'trigger': 'stopped',
            'source': '*',
            'dest': 'loaded',
            'before': 'trigger_stopped_events',
            'conditions': ['is_in_use']
        }
    ]

    def __init__(self, filename, mapping, name=None, gain=1):
        Machine(model=self, states=self.STATES,
                transitions=self.TRANSITIONS, initial='initial',
                ignore_invalid_triggers=True)

        self.loaded_callbacks = []
        self.mapping = mapping
        self.filename = filename
        self.name = name or filename
        self.audio_segment = None
        self.initial_volume_factor = gain
        self.music_lock = Lock("music__" + filename)

        threading.Thread(name="MSMusicLoad", target=self.load).start()

    def reload_properties(self, name=None, gain=1):
        self.name = name or self.filename
        if gain != self.initial_volume_factor:
            self.initial_volume_factor = gain
            self.reload_music_file()

    def reload_music_file(self):
        with file_lock:
            try:
                if self.filename.startswith("/"):
                    filename = self.filename
                else:
                    filename = Config.music_path + self.filename

                debug_print("Reloading « {} »".format(self.name))
                initial_db_gain = gain(self.initial_volume_factor * 100)
                self.audio_segment = pydub.AudioSegment \
                        .from_file(filename) \
                        .set_frame_rate(Config.frame_rate) \
                        .set_channels(Config.channels) \
                        .set_sample_width(Config.sample_width) \
                        .apply_gain(initial_db_gain)
            except Exception as e:
                error_print("failed to reload « {} »: {}"\
                        .format(self.name, e))
                self.loading_error = e
                self.to_failed()
            else:
                debug_print("Reloaded « {} »".format(self.name))

    # Machine related events
    def on_enter_loading(self):
        with file_lock:
            try:
                if self.filename.startswith("/"):
                    filename = self.filename
                else:
                    filename = Config.music_path + self.filename

                debug_print("Loading « {} »".format(self.name))
                self.mixer = self.mapping.mixer or Mixer()
                initial_db_gain = gain(self.initial_volume_factor * 100)
                self.audio_segment = pydub.AudioSegment \
                        .from_file(filename) \
                        .set_frame_rate(Config.frame_rate) \
                        .set_channels(Config.channels) \
                        .set_sample_width(Config.sample_width) \
                        .apply_gain(initial_db_gain)
                self.sound_duration = self.audio_segment.duration_seconds
            except Exception as e:
                error_print("failed to load « {} »: {}".format(self.name, e))
                self.loading_error = e
                self.fail()
            else:
                self.success()
                debug_print("Loaded « {} »".format(self.name))

    def on_enter_loaded(self):
        self.cleanup()

    def cleanup(self):
        self.gain_effects = []
        self.set_gain(0, absolute=True)
        self.current_audio_segment = None
        self.volume = 100
        self.wait_event = threading.Event()
        self.current_loop = 0

    def on_enter_loaded_playing(self):
        self.mixer.add_file(self)

    # Machine related states
    def is_in_use(self):
        return self.is_loaded(allow_substates=True) and not self.is_loaded()

    def is_in_use_not_stopping(self):
        return self.is_loaded_playing() or self.is_loaded_paused()

    # Machine related triggers
    def trigger_stopped_events(self):
        self.mixer.remove_file(self)
        self.wait_event.set()
        self.cleanup()

    # Actions and properties called externally
    @property
    def sound_position(self):
        if self.is_in_use():
            return self.current_frame / self.current_audio_segment.frame_rate
        else:
            return 0

    def play(self, fade_in=0, volume=100, loop=0, start_at=0):
        self.set_gain(gain(volume) + self.mapping.master_gain, absolute=True)
        self.volume = volume
        if loop < 0:
            self.last_loop = float('inf')
        else:
            self.last_loop = loop

        with self.music_lock:
            self.current_audio_segment = self.audio_segment
            self.current_frame = int(start_at * self.audio_segment.frame_rate)

        self.start_playing()

        if fade_in > 0:
            db_gain = gain(self.volume, 0)[0]
            self.set_gain(-db_gain)
            self.add_fade_effect(db_gain, fade_in)

    def seek(self, value=0, delta=False):
        if not self.is_in_use_not_stopping():
            return

        with self.music_lock:
            self.abandon_all_effects()
            if delta:
                frame_count = int(self.audio_segment.frame_count())
                frame_diff = int(value * self.audio_segment.frame_rate)
                self.current_frame += frame_diff
                while self.current_frame < 0:
                    self.current_loop -= 1
                    self.current_frame += frame_count
                while self.current_frame > frame_count:
                    self.current_loop += 1
                    self.current_frame -= frame_count
                if self.current_loop < 0:
                    self.current_loop = 0
                    self.current_frame = 0
                if self.current_loop > self.last_loop:
                    self.current_loop = self.last_loop
                    self.current_frame = frame_count
            else:
                self.current_frame = max(
                        0,
                        int(value * self.audio_segment.frame_rate))

    def stop(self, fade_out=0, wait=False, set_wait_id=None):
        if self.is_loaded_playing():
            ms = int(self.sound_position * 1000)
            ms_fo = max(1, int(fade_out * 1000))

            new_audio_segment = self.current_audio_segment[: ms+ms_fo] \
                .fade_out(ms_fo)
            with self.music_lock:
                self.current_audio_segment = new_audio_segment
            self.stop_playing()
            if wait:
                self.mapping.add_wait(self.wait_event, wait_id=set_wait_id)
                self.wait_end()
        else:
            self.stopped()

    def abandon_all_effects(self):
        db_gain = 0
        for gain_effect in self.gain_effects:
            db_gain += gain_effect.get_last_gain()

        self.gain_effects = []
        self.set_gain(db_gain)

    def set_volume(self, value, delta=False, fade=0):
        [db_gain, self.volume] = gain(
                value + int(delta) * self.volume,
                self.volume)

        self.set_gain_with_effect(db_gain, fade=fade)

    def set_gain_with_effect(self, db_gain, fade=0):
        if not self.is_in_use():
            return

        if fade > 0:
            self.add_fade_effect(db_gain, fade)
        else:
            self.set_gain(db_gain)

    def wait_end(self):
        self.wait_event.clear()
        self.wait_event.wait()

    # Let other subscribe for an event when they are ready
    def subscribe_loaded(self, callback):
        # FIXME: should lock to be sure we have no race, but it makes the
        # initialization screen not showing until everything is loaded
        if self.is_loaded(allow_substates=True):
            callback(True)
        elif self.is_failed():
            callback(False)
        else:
            self.loaded_callbacks.append(callback)

    def poll_loaded(self):
        for callback in self.loaded_callbacks:
            callback(self.is_loaded())
        self.loaded_callbacks = []

    # Callbacks
    def finished_callback(self):
        self.stopped()

    def play_callback(self, out_data_length, frame_count):
        if self.is_loaded_paused():
            return b'\0' * out_data_length

        with self.music_lock:
            [data, nb_frames] = self.get_next_sample(frame_count)
            if nb_frames < frame_count:
                if self.is_loaded_playing() and\
                        self.current_loop < self.last_loop:
                    self.current_loop += 1
                    self.current_frame = 0
                    [new_data, new_nb_frames] = self.get_next_sample(
                            frame_count - nb_frames)
                    data += new_data
                    nb_frames += new_nb_frames
                elif nb_frames == 0:
                    # FIXME: too slow when mixing multiple streams
                    threading.Thread(
                            name="MSFinishedCallback",
                            target=self.finished_callback).start()

            return data.ljust(out_data_length, b'\0')

    # Helpers
    def set_gain(self, db_gain, absolute=False):
        if absolute:
            self.db_gain = db_gain
        else:
            self.db_gain += db_gain

    def get_next_sample(self, frame_count):
        fw = self.audio_segment.frame_width

        data = b""
        nb_frames = 0

        segment = self.current_audio_segment
        max_val = int(segment.frame_count())

        start_i = max(self.current_frame, 0)
        end_i   = min(self.current_frame + frame_count, max_val)
        data += segment._data[start_i*fw : end_i*fw]
        nb_frames += end_i - start_i
        self.current_frame += end_i - start_i

        volume_factor = self.volume_factor(self.effects_next_gain(nb_frames))

        data = audioop.mul(data, Config.sample_width, volume_factor)

        return [data, nb_frames]

    def add_fade_effect(self, db_gain, fade_duration):
        if not self.is_in_use():
            return

        self.gain_effects.append(GainEffect(
            "fade",
            self.current_audio_segment,
            self.current_loop,
            self.sound_position,
            self.sound_position + fade_duration,
            gain=db_gain))

    def effects_next_gain(self, frame_count):
        db_gain = 0
        for gain_effect in self.gain_effects:
            [new_gain, last_gain] = gain_effect.get_next_gain(
                    self.current_frame,
                    self.current_loop,
                    frame_count)
            if last_gain:
                self.set_gain(new_gain)
                self.gain_effects.remove(gain_effect)
            else:
                db_gain += new_gain
        return db_gain


    def volume_factor(self, additional_gain=0):
        return 10 ** ( (self.db_gain + additional_gain) / 20)

